#!/usr/bin/env python
#
#       collision.py
#       
#       Copyright 2009 Gordon Allott <gordallott@gmail.com>
#       
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 3 of the License, or
#       (at your option) any later version.
#       
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#       
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.

""" We use pymunk (python chipmunk) to do our collision detection, this is 
    Mostly for speed reasons, I think i could get a collision detection engine
    running for our needs thats slightly slower but without the external dependancy
    and if i have time later I will, but for now this is the right way
"""

import ctypes as ct
import pymunk as pm
from pymunk import Vec2d
import math, sys, random

COLLISION_TYPE_DEFAULT = 0x0000
COLLISION_TYPE_GROUP1 = 0x0001
COLLISION_TYPE_GROUP2 = 0x0002

def cpvrotate(v1, v2):
    return vec2d(v1.x*v2.x - v1.y*v2.y, v1.x*v2.y + v1.y*v2.x)
def cpvadd(v1, v2):
    return vec2d(v1.x + v2.x, v1.y + v2.y)
def cpvzero():
    return vec2d(0,0)
def flipy(y):
    return -y+600
    
#a lot of this comes from the pymunk example code - im not 100% familar with 
#chipmunk's best practices

def callback_collision_generic(shapeA, shapeB, contacts, normal, data):
    #print 'Collision!', shapeA.entity, '-', shapeB.entity
    return True

def callback_collision_mouse(shape, data):
    print 'mouse collided with shape', shape.entity

#def coll_func(shape, screen):
    #"""Draw a circle at the contact"""
    #if hasattr(shape, "radius"):
        #r = shape.radius + 4
    #else:
        #r = 10
    #p = to_pygame(shape.body.position)
    #pygame.draw.circle(screen, THECOLORS["red"], p, r, 2)

class physworld(object):
    
    _inited = False
    _space = None
    _collisions = None
    
    def __getattr__(self, attribute):
        if attribute == 'inited':
            return self.__class__._inited
        elif attribute == 'space':
            return self.__class__._space
        elif attribute == 'collisions':
            return self.__class__._collisions
    
    def __setattr__(self, attribute, value):
        if attribute == 'inited':
            self.__class__._inited = value
        elif attribute == 'space':
            self.__class__._space = value
        elif attribute == 'collisions':
            self.__class__._collisions = value
        else:
            self.__dict__[attribute] = value
    
    def __init__(self):
        if not self.inited:
            pm.init_pymunk()
            self.inited = True
            self.collisions = set()
            
        if not self.space:            
            pm.init_pymunk()
            self.space = pm.Space()
            self.space.gravity = Vec2d(0.0, 0.0)
                
            self.space.resize_static_hash()
            self.space.resize_active_hash()
            
            self.space.set_default_collisionpair_func(self.callback_collision_generic)
            
        self.mouse_entity = None
            
    def add(self, ent):
        for shape in ent.shapes:
            self.space.add(shape)
            
    def tick(self):
        #we use a set() so that only one of each collision is reported
        self.collisions = set()
        self.space.step(1.0/30)

        #we should now have a bunch of collisions in self.collisions, 
        # we wait until here to pass that data on so that the AI does not
        # destroy/move things whist in the step() function
        for collision in self.collisions:
            entityA, entityB = collision
            getattr(entityA.ai, 'collision_with_%s' % (entityB.type))(entityB)
        
    def mouse_update(self, mouse_pos):#, callback=None):
        callback = self.callback_collision_mouse
        mouse_pos = Vec2d(mouse_pos[0], mouse_pos[1])
        
        self.mouse_set = set()
        self.space.nonstatic_point_query(mouse_pos, callback, None)
        
        def sortfunc(a, b):
            return int(b.cached_z - a.cached_z)
            
        self.mouse_set = sorted(self.mouse_set, sortfunc)
        
        delete = []
        for item in self.mouse_set:
            if not item.mouse_sensitive:
                delete.append(item)
                
        for item in delete:
            self.mouse_set.pop(item)
        
        if len(self.mouse_set):
            self.mouse_entity = self.mouse_set[0]
        else:
            self.mouse_entity = None
        
        
    def callback_collision_generic(self, shapeA, shapeB, contacts, normal, data):

        if hasattr(shapeA.entity.ai, 'collision_with_%s' % (shapeB.entity.type)):
            #we have ShapeA.entity.ai.collision_with_type()
            self.collisions.add((shapeA.entity, shapeB.entity))
            
        if hasattr(shapeB.entity.ai, 'collision_with_%s' % (shapeA.entity.type)):
            self.collisions.add((shapeB.entity, shapeA.entity))
        
        return True

    def callback_collision_mouse(self, shape, data):
        if shape.entity.mouse_sensitive:
            self.mouse_set.add(shape.entity)
            
            
    def remove_shape(self, shape):
        print 'removing shape', shape
        self.space.remove(shape)
        
    

        

        
        
class Shape(object):
    
    
    #__slots__ = ['body', 'shapes', 'parent', 'collision_type']
    body = None
    shapes = []
    collision_type = COLLISION_TYPE_DEFAULT
    
    parent = None
    
    def __init__(self, parent, collision_type=COLLISION_TYPE_DEFAULT):
        # PARENT MUST BE SUPPLIED, its essentially the entity object that
        # called this one, its *important*
        self.__class__.physworld = physworld()
        self.parent = parent
        self.shapes = []
        self.body = None
        
        self.collision_type = collision_type
        
    def new_body(self, position=(0.0, 0.0)):
        self.body = pm.Body(pm.inf, pm.inf)
        self.body.position = Vec2d(position)
        
    def _add_shape(self, shape):
        shape.collision_type = self.collision_type
        shape.entity = self.parent
        self.shapes += [shape] #add the shape to our list of shapes
        
        #cp.cpSpaceAddShape(self.world.space, shape)
        #cp.cpShapeCacheBB(shape)

    def __del__(self):
        for shape in self.shapes:
            self.__class__.physworld.remove_shape(shape)
            
        
    def new_shape_circle(self, radius, position=(0.0, 0.0)):
        shape = pm.Circle(self.body, radius, Vec2d(position))
        self._add_shape(shape)


    def new_shape_poly(self, points):
        poly_points = pm.util.convex_hull(points)
        
        if not pm.util.is_clockwise(poly_points):
            poly_points.reverse()  
        
        # Create Shape
        shape = pm.Poly(self.body, poly_points, Vec2d(0,0))
        self._add_shape(shape)
        
    def set_position(self, position):
        self.body.position = Vec2d(position)

        
        # Change vectors to the point of view of the center
        #poly_points_center = pm.util.poly_vectors_around_center(poly_points)
        
        
        
    #def new_shape_segments(self, vertices, thickness=0.0):
        #shapes = []

        #for index in range(len(vertices) - 1):
            #v1 = vec2d(vertices[index])
            #v2 = vec2d(vertices[index+1])
            
            #shape = cp.cpSegmentShapeNew(self.body, v1, v2, thickness)
            #shape.contents.u = self.friction
            #print 'foo', ct.cast(shape, ct.POINTER(cp.cpSegmentShape)).contents.a, ct.cast(shape, ct.POINTER(cp.cpSegmentShape)).contents.b
            ##shape.contents.e = self.springyness
            #shape.collison_type = self.collision_type
            
            #self._add_shape(shape)
            #print 'foo', ct.cast(shape, ct.POINTER(cp.cpSegmentShape)).contents.a, ct.cast(shape, ct.P
        
        
