#!/usr/bin/env python
# -*- coding: utf-8 -*-
#       timeline.py
#
#       Copyright 2009 Gordon Allott <gordallott@gmail.com>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
#
#       #

""" This is basically all copied from CSL-I from the last pyweek, modified to use
SDL instead of pygame and some optimisations so i can instance it anywhere 
"""

from SDL import *
import random


time_now = 0
time_start = 1

def make_id(store=None):

    if not store:
        return random.randrange(0, 2**16)

    id = 0
    while True:
        if not store.has_key(id):
            break

        else:
            id = random.randrange(0, 2**16)

    return id


class Timeline:

    inited = False

    _start_time = None
    _current_time = None

    _actions = None
    _pre_delete = None

    def __getattr__(self, attribute):
        if attribute == 'start_time':
            return self.__class__._start_time
        elif attribute == 'current_time':
            return self.__class__._current_time
        elif attribute == 'actions':
            return self.__class__._actions
        elif attribute == 'pre_delete':
            return self.__class__._pre_delete

        
        
    def __setattr__(self, attribute, value):
        if attribute == 'start_time':
            self.__class__._start_time = value
        elif attribute == 'current_time':
            self.__class__._current_time = value
        elif attribute == 'actions':
            self.__class__._actions = value
        elif attribute == 'pre_delete':
            self.__class__._pre_delete = value
        else:
            self.__dict__[attribute] = value

    def __init__(self):
        if not self.__class__.inited:
            print 'timeline init'
            self.actions = {}
            self.start_time = SDL_GetTicks()
            self.pre_delete = []
            self.current_time = self.start_time
            self.step()
            
            print self.current_time
            
            self.__class__.inited = True

    def remove_action(self, action_id, finish=True):
        self.pre_delete.append({'id':action_id, 'finish':finish})

    def step(self):
        clock_tick = SDL_GetTicks()
        self.current_time += clock_tick

        ctime = self.current_time / 1000.0

        delete_candidates = []
        
        for action in self.pre_delete:
            target_action = self.actions[action['id']]
            if action['finish']:
                if target_action['repeat_call']:
                    target_action['callback'](1.0, **target_action['extra_args'])
            self.actions.pop(action['id'])

        for id, action in self.actions.iteritems():
            if self.check_action(action, ctime):
                action['callback'](self.get_progress(ctime, action['start'], action['length']), **action['extra_args'])
                action['iterations'] += 1

            if self.check_action_deletion(action, ctime):
                delete_candidates.append(id)


        for id in delete_candidates:
            #fire off one last callback if its a repeat_callback with a definate
            #progress of 1
            if self.actions[id]['repeat_call']:
                self.actions[id]['callback'](1.0, **self.actions[id]['extra_args'])
            self.actions.pop(id)
        
        self.pre_delete = []

    def get_progress(self, ctime, stime, length):
        if length == 0.0:
            return 1.0

        ptime = ctime - stime
        return min(ptime / length, 1.0)

    def check_action_deletion(self, action, ctime):

        if not action['repeating']:
            if not(action['repeat_call']) and action['iterations'] > 0:
                return True

            if action['start'] + action['length'] < ctime:
                return True

        else:
            # we have a repeating action,
            taction = action
            taction['repeating'] = False
            do_delete = self.check_action_deletion(taction, ctime)
            taction['repeating'] = True
            if do_delete:
                action['start'] = ctime
                action['iterations'] = 0

        return False


    def check_action(self, action, ctime):

        if ctime < action['start']:
            return False

        #check non repeat call actions first
        if not action['repeat_call']:
            if action['iterations'] > 0:
                return False

            return True

        else:
            if action['start'] - ctime < action['length']:
                return True

            return False

        return True



    def add_action(self, time=0.0, relative=time_start, callback=None,
                   repeat_action=False, repeat_callback=False, length=0.0,
                   wait=0.0, extra_args={}):
        """ in time seconds relative to relative the callback callback will be
        called
        """

        base_action = {
            'start': 0.0,
            'length': 0.0,
            'repeating': False,
            'repeat_call': False,
            'repeat_wait': 0.0,
            'callback': None,
            'iterations': 0
        }

        start = 0.0 if relative == time_start else self.current_time / 1000.0
        start += time

        length = length
        repeating = repeat_action
        repeat_call = repeat_callback
        repeat_wait = wait

        action = base_action
        action['start'] = start
        action['length'] = length
        action['repeating'] = repeat_action
        action['repeat_call'] = repeat_call
        action['repeat_wait'] = repeat_wait
        action['callback'] = callback
        action['extra_args'] = extra_args

        id = make_id(self.actions)

        self.actions[id] = action
        return id

    #def remove_action(self, id):
        #self.actions.pop(id)
